
from odoo import api, fields, models, _
from odoo.osv import expression
from odoo.tools import groupby


class DeferredReportCustomHandler(models.AbstractModel):
    _inherit = 'account.deferred.report.handler'

    def _get_custom_display_config(self):
        return {
            'templates': {
                'AccountReportFilters': 'openerp_enterprise.DisallowedExpensesFleetReportFilters',
            },
        }

    def _custom_options_initializer(self, report, options, previous_options=None):
        super()._custom_options_initializer(report, options, previous_options=previous_options)

        # initialize hosting_split filter (disabled by default)
        options['hosting_split'] = previous_options.get('hosting_split', False)
        # intialize commission_plan_split filter (disabled by default)
        options['commission_plan_split'] = previous_options.get('commission_plan_split', False)

    def _get_domain(self, report, options, filter_already_generated=False, filter_not_started=False):
        domain = super()._get_domain(
            report, options,
            filter_already_generated=filter_already_generated, filter_not_started=filter_not_started
        )
        if '__domain_filter_hosting' in self.env.context:
            # for cell audit, add filter by hosting when clicking on a splitted hosting line
            domain = expression.AND((
                domain,
                [('hosting', '=', self.env.context['__domain_filter_hosting'])]
            ))
        if '__domain_filter_commission' in self.env.context:
            commission_id = self.env.context['__domain_filter_commission']
            if commission_id:
                commission_domain = [
                    '|', ('subscription_id.commission_plan_id', '=', commission_id),
                         ('sale_line_ids.order_id.commission_plan_id', '=', commission_id),
                ]
            else:
                commission_domain = [
                    '&', ('subscription_id.commission_plan_id', '=', commission_id),
                    '!', ('sale_line_ids.order_id.commission_plan_id', '!=', False),
                ]
            domain = expression.AND((domain, commission_domain))
        return domain

    @api.model
    def _get_select(self):
        return super()._get_select() + [
            "account_move_line.hosting AS hosting",
            """
            COALESCE(
                account_move_line__comm_subscription_id.commission_plan_id,
                account_move_line__comm_order_id.commission_plan_id
            ) AS commission_plan_id
            """.strip(),
        ]

    def _query_get_extra_from(self):
        """return extra query FROM clause needed to compute commercial_plan_id"""
        return """
            LEFT JOIN "sale_order" AS "account_move_line__comm_subscription_id"
                   ON ("account_move_line"."subscription_id" = "account_move_line__comm_subscription_id"."id")
            LEFT JOIN LATERAL (
                -- get the commission plan of the most recent order (date_order) related to the move line
                   SELECT (array_agg(so.commission_plan_id))[1] AS commission_plan_id
                     FROM sale_order_line_invoice_rel solinv_rel
                LEFT JOIN sale_order_line sol ON (solinv_rel.order_line_id = sol.id)
                     JOIN LATERAL (
                        -- get orders related to the move line ordered by order date (most recent first)
                        SELECT so2.commission_plan_id
                          FROM sale_order so2
                         WHERE so2.id = sol.order_id
                      ORDER BY so2.date_order DESC, so2.id DESC
                     ) AS so ON TRUE
                WHERE solinv_rel.invoice_line_id = account_move_line.id
                GROUP BY solinv_rel.invoice_line_id
            ) AS "account_move_line__comm_order_id" ON TRUE
        """

    def _get_lines(self, report, options, filter_already_generated=False):
        """inherited to add extra tables in FROM needed to compute commission_plan_id."""
        domain = self._get_domain(report, options, filter_already_generated)
        tables, where_clause, where_params = report._query_get(options, domain=domain, date_scope='from_beginning')
        select_clause = ', '.join(self._get_select())

        # NOTE(inherit): manually add join for tables needed to get commission plan
        # (as that information is not stored on the move line itself)
        tables += self._query_get_extra_from()

        query = f"""
        SELECT {select_clause}
        FROM {tables}
        LEFT JOIN product_product AS account_move_line__product_id ON account_move_line.product_id = account_move_line__product_id.id
        LEFT JOIN product_template AS account_move_line__product_template_id ON account_move_line__product_id.product_tmpl_id = account_move_line__product_template_id.id
        WHERE {where_clause}
        ORDER BY "account_move_line"."deferred_start_date", "account_move_line"."id"
        """

        self.env.cr.execute(query, where_params)
        res = self.env.cr.dictfetchall()
        return res

    def _get_hosting_selection_values(self):
        return self.env["account.move.line"]._fields['hosting']._description_selection(self.env) + [
            (False, _("(not set)")),  # always return False last
        ]

    def action_audit_cell(self, options, params):
        hosting_selvals = self._get_hosting_selection_values()
        report = self.env['account.report'].browse(options['report_id'])
        markup, model, value = report._parse_line_id(params.get('calling_line_dict_id'))[-1]
        while markup == "hosting" or model == "commission.plan":
            if not model:  # hosting split has no model
                try:
                    hosting = hosting_selvals[int(value) or 0][0]
                    # set context so that we add hosting filter in `get_domain()`
                    self = self.with_context(__domain_filter_hosting=hosting)
                except IndexError:
                    pass  # wrong hosting index, ignore
            elif model == "commission.plan":
                self = self.with_context(__domain_filter_commission=value)
            # update line id to match the parent, to ensure we filter on the account
            params['calling_line_dict_id'] = (
                report._build_parent_line_id(report._parse_line_id(params['calling_line_dict_id']))
            )
            markup, model, value = report._parse_line_id(params.get('calling_line_dict_id'))[-1]
        return super().action_audit_cell(options, params)

    def _dynamic_lines_generator(self, report, options, all_column_groups_expression_totals, warnings=None):
        """inherited to dynamically add 'Hosting'/'Commission Plan' lines as child of 'Account' line (if option enabled)

        NOTE: to avoid computing the amounts twice, this method is most a copy of the same method
          from account_reports/models/account_deferred_reports.py with additions to generate lines
          for "Hosting"/"Commission Plan" splits (see "NOTE(inherit)" comments)
        """
        def get_columns(totals):
            return [
                {
                    **report._build_column_dict(
                        totals[(
                            fields.Date.to_date(column['date_from']),
                            fields.Date.to_date(column['date_to']),
                            column['expression_label']
                        )],
                        column,
                        options=options,
                        currency=self.env.company.currency_id,
                    ),
                    'auditable': True,
                }
                for column in options['columns']
            ]

        lines = self._get_lines(report, options)
        periods = [
            (
                fields.Date.from_string(column['date_from']),
                fields.Date.from_string(column['date_to']),
                column['expression_label'],
            )
            for column in options['columns']
        ]

        deferred_amounts_by_line = self.env['account.move']._get_deferred_amounts_by_line(lines, periods)
        totals_per_grouping_field, totals_all_grouping_field = self._group_deferred_amounts_by_grouping_field(
            deferred_amounts_by_line=deferred_amounts_by_line,
            periods=periods,
            is_reverse=self._get_deferred_report_type() == 'expense',
            filter_already_generated=False,
            grouping_field=options['deferred_grouping_field'],
        )

        report_lines = []

        grouping_model = self.env['account.move.line'][options['deferred_grouping_field']]._name
        for totals_grouping_field in totals_per_grouping_field.values():
            grouping_record = self.env[grouping_model].browse(totals_grouping_field[options['deferred_grouping_field']])
            grouping_field_description = self.env['account.move.line'][options['deferred_grouping_field']]._description
            if options['deferred_grouping_field'] == 'product_id':
                grouping_field_description = _("Product")
            grouping_name = grouping_record.display_name or _("(No %s)", grouping_field_description)
            report_lines.append((0, {
                'id': report._get_generic_line_id(grouping_model, grouping_record.id),
                'name': grouping_name,
                'caret_options': 'deferred_caret',
                'level': 1,
                'columns': get_columns(totals_grouping_field),
            }))

            # NOTE(inherit): depending on the enabled options, append the splitted amount
            # by split options as children of the account line.
            #  As we have no way to known in with order the options were enabled by the
            #  user on the report, we always group by hosting first, then by commission plan.
            account_subgroups = []
            if options.get('hosting_split'):
                account_subgroups.append("hosting")
            if options.get("commission_plan_split"):
                account_subgroups.append("commission_plan_id")
            report_lines += self._get_account_subgroups_lines(
                report, account_subgroups, totals_grouping_field, report_lines[-1][1], get_columns,
            )

        if totals_per_grouping_field:
            report_lines.append((0, {
                'id': report._get_generic_line_id(None, None, markup='total'),
                'name': 'Total',
                'level': 1,
                'columns': get_columns(totals_all_grouping_field),
            }))

        return report_lines

    def _get_account_subgroups_lines(self, report, subgroups, totals, parent_line, get_columns_func):
        """Return report lines for subgroups (Hosting/Commision Plan) of account line."""
        if not subgroups:
            return []
        report_lines = []
        subgroup, next_subgroups = subgroups[0], subgroups[1:]
        subgroup_totals = totals.get(f"subgroup_{subgroup}", {})
        if subgroup == "hosting":
            # NOTE: precompute the hosting selection {value: label} dict
            #  and the order of those values (used for report line id)
            hostings = self._get_hosting_selection_values()
            hosting_names = dict(hostings)
            hosting_name_order = {k: i for i, (k, v) in enumerate(hostings)}
            for hosting, totals_hosting in sorted(subgroup_totals.items(),
                                                  key=lambda h: hosting_name_order[h[0]]):
                report_lines.append((0, {
                    'id': report._get_generic_line_id(
                        # As report's frontend only support integer for line id sub. part,
                        # use the index of the value in the hosting selection instead
                        # of the value itself (it will be used for ex. when clicking
                        # on a cell to determine the hosting filter to apply, see
                        # `action_audit_cell()` method)
                        "", hosting_name_order[hosting],
                        markup="hosting",
                        parent_line_id=parent_line['id'],
                    ),
                    'parent_id': parent_line['id'],
                    'name': _("Hosting: %s", hosting_names[hosting]),
                    'caret_options': False,
                    'unfoldable': False,
                    'level': parent_line['level'] + 1,
                    'columns': get_columns_func(totals_hosting),
                }))
                report_lines += self._get_account_subgroups_lines(
                    report, next_subgroups, totals_hosting, report_lines[-1][1], get_columns_func,
                )

        elif subgroup == "commission_plan_id":
            for commission_plan_id, totals_commission in subgroup_totals.items():
                commission_plan = self.env["commission.plan"].browse(commission_plan_id)
                commission_plan_name = commission_plan.name if commission_plan else _('(not set)')
                report_lines.append((0, {
                    'id': report._get_generic_line_id(
                        "commission.plan", commission_plan_id,
                        parent_line_id=parent_line['id'],
                    ),
                    'parent_id': parent_line['id'],
                    'name': _("Comm. Plan: %s", commission_plan_name),
                    'caret_options': False,
                    'unfoldable': False,
                    'level': parent_line['level'] + 1,
                    'columns': get_columns_func(totals_commission),
                }))
                report_lines += self._get_account_subgroups_lines(
                    report, next_subgroups, totals_commission, report_lines[-1][1], get_columns_func,
                )
        return report_lines

    @api.model
    def _group_deferred_amounts_by_grouping_field(self, deferred_amounts_by_line, periods, is_reverse, filter_already_generated=False, grouping_field='account_id'):
        """
        Groups the deferred amounts by account and computes the totals for each account for each period.
        And the total for all accounts for each period.
        E.g. (where period1 = (date1, date2, label1), period2 = (date2, date3, label2), ...)
        {
            self._get_grouping_keys_deferred_lines(): {
                'account_id': account1, 'amount_total': 600, period_1: 200, period_2: 400,
                'hosting': {
                    'online': {'amount_total': 300, period_1: 100, period_2: 200},
                    'paas': {'amount_total': 200, period_1: 50, period_2: 150},
                    'premise': {'amount_total': 100, period_1: 50, period_2: 50},
            },
            self._get_grouping_keys_deferred_lines(): {
                'account_id': account2, 'amount_total': 700, period_1: 300, period_2: 400
                'hosting:' {
                    False: {'amount_total': 700, period_1: 300, period_2: 400}
                }
            },
        }, {'totals_aggregated': 1300, period_1: 500, period_2: 800}

        Inherited to add amounts splitted by hosting
        """
        deferred_amounts_by_line = groupby(deferred_amounts_by_line,
                                           key=lambda x: self._group_by_deferred_fields(x, filter_already_generated, grouping_field))
        totals_per_key = {}  # {key: {**self._get_grouping_fields_deferral_lines(), total, before, current, later}}
        totals_aggregated_by_period = {period: 0 for period in periods + ['totals_aggregated']}
        sign = 1 if is_reverse else -1
        for key, lines_per_key in deferred_amounts_by_line:
            lines_per_key = list(lines_per_key)
            current_key_totals = self._get_current_key_totals_dict(lines_per_key, sign)
            totals_aggregated_by_period['totals_aggregated'] += current_key_totals['amount_total']
            for period in periods:
                current_key_totals[period] = sign * sum(line[period] for line in lines_per_key)
                totals_aggregated_by_period[period] += self.env.company.currency_id.round(current_key_totals[period])
            totals_per_key[key] = current_key_totals

            # recursively group lines by fields of each custom subgroup
            # Note: group key is prefixed by `subgroup_` to avoid conflicts with existing keys.
            def _group_totals_by_fields(line_dict, lines, group_fields):
                field, next_fields = group_fields[0], group_fields[1:]
                group_key = f"subgroup_{field}"
                line_dict.setdefault(group_key, {})
                line_group = line_dict[group_key]
                group_by_field = groupby(lines, key=lambda x: x.get(field) or False)
                for field_value, lines_per_field in group_by_field:
                    lines_per_field = list(lines_per_field)
                    line_group.setdefault(field_value, {period: 0 for period in periods + ['amount_total']})
                    line_group[field_value]['amount_total'] += sign * sum(line['balance'] for line in lines_per_field)
                    for period in periods:
                        line_group[field_value][period] += sign * sum(line[period] for line in lines_per_field)
                    if next_fields:
                        _group_totals_by_fields(line_group[field_value], lines_per_field, next_fields)

            subgroups = [
                ('hosting', 'commission_plan_id'),
                ('commission_plan_id',),
            ]
            for subgroup_fields in subgroups:
                _group_totals_by_fields(current_key_totals, lines_per_key, subgroup_fields)

        return totals_per_key, totals_aggregated_by_period

    def open_journal_items(self, options, params):
        action = super().open_journal_items(options, params)
        if options.get('hosting_split'):
            # if split by hosting is enabled, group by it when opening journal items
            # for a specific account
            action['context']['search_default_group_by_hosting'] = 1
        return action
